import 'dart:io';

import 'package:markdown/markdown.dart';
import 'package:path/path.dart' as path;

import 'src/template.dart';

class Documentation {
  const Documentation({
    required this.name,
    required this.root,
  });

  final String name;
  final Page root;

  String render() {
    final menu = '<ul>${root.renderMenu(0)}</ul>';
    final content = root.renderContent();
    return pageTemplate(name, menu, content);
  }

  static Future<Documentation> load(String name, Directory directory) async {
    return Documentation(
      name: name,
      root: (await Page.load(directory, name)),
    );
  }
}

class Page {
  const Page({
    required this.name,
    required this.path,
    required this.content,
    required this.children,
  });

  Page withName(String name) => Page(
        name: name,
        path: path,
        content: content,
        children: children,
      );

  String get id => path.isEmpty ? '-home' : Uri.encodeComponent(path.join('-'));
  final String name;
  final List<String> path;
  final String content;
  final List<Page> children;

  String renderMenu(int scope) {
    final buffer = StringBuffer();
    if (scope > 0) {
      buffer.write(
          '<li class="${scope <= 1 ? 'primary' : 'secondary'}"><a id="menu--$id" href="#$id">$name</a>');
    }
    if (children.isNotEmpty) {
      if (scope > 0) buffer.write('<ul>');
      for (var item in children) {
        buffer.write(item.renderMenu(scope + 1));
      }
      if (scope > 0) buffer.write('</ul>');
    }
    if (scope > 0) {
      buffer.write('</li>');
    }
    return buffer.toString();
  }

  String renderContent() {
    final buffer = StringBuffer(
        '<div id="content-$id" class="page" style="display:none">$content</div>');

    for (var item in children) {
      buffer.write(item.renderContent());
    }

    return buffer.toString();
  }

  static Future<Page> load(Directory directory, String name) {
    return _loadDirectory(
      [],
      name,
      directory,
    );
  }
}

Future<Page> _loadDirectory(
    List<String> currentPath, String name, Directory directory) async {
  final index = File('index.md');
  final hasContent = index.existsSync();
  final children = <Page>[];
  for (var item in await directory.list().toList()) {
    if (item is Directory) {
      final hasMdFiles = await directory
          .list(recursive: true)
          .any((x) => path.extension(x.path) == '.md');
      if (hasMdFiles) {
        final itemName = path.basenameWithoutExtension(item.path);
        children.add(await _loadDirectory(<String>[
          ...currentPath,
          itemName,
        ], itemName, item));
      }
    } else if (item is File) {
      if (path.extension(item.path) == '.md' &&
          path.basename(item.path).toLowerCase() != 'index.md') {
        children.add(await _loadFile(<String>[
          ...currentPath,
          path.basenameWithoutExtension(item.path),
        ], item));
      }
    }
  }
  final content = hasContent
      ? await index.readAsString()
      : '# $name\n\n${children.map((x) => '* [${x.name}](#${x.id})').join('\n')}';

  return Page(
    name: name,
    path: currentPath,
    content: renderMarkdown(content),
    children: children,
  );
}

Future<Page> _loadFile(List<String> currentPath, File file) async {
  final name = path.basenameWithoutExtension(file.path);
  final content = await file.readAsString();

  return Page(
    name: name,
    path: currentPath,
    content: renderMarkdown(content),
    children: const <Page>[],
  );
}

String renderMarkdown(
  String markdown, {
  Iterable<BlockSyntax> blockSyntaxes = const [],
  Iterable<InlineSyntax> inlineSyntaxes = const [],
  ExtensionSet? extensionSet,
  Resolver? linkResolver,
  Resolver? imageLinkResolver,
  bool inlineOnly = false,
}) {
  var document = Document(
    blockSyntaxes: blockSyntaxes,
    inlineSyntaxes: inlineSyntaxes,
    extensionSet: extensionSet,
    linkResolver: linkResolver,
    imageLinkResolver: imageLinkResolver,
  );

  // Replace windows line endings with unix line endings, and split.
  var lines = markdown.replaceAll('\r\n', '\n').split('\n');
  final nodes = document.parseLines(lines);

  // TODO replace relative path

  return renderToHtml(nodes) + '\n';
}
